#include <MFRC522Extended.h>
#include <SPI.h>
#include <MFRC522.h>

const int BUTTON_PIN = 2;
const int SS_PIN = 10;
const int RST_PIN = 5;
const int BUZZER_PIN = 9;
const int ECHO_PIN = 8;

volatile bool lockState = false;
volatile bool buttonPressed = false;
volatile unsigned long buttonLastStateTime = 0;

byte keyTagUID[4];
bool withValidKeycard = false;
unsigned long duration, distance;
byte bufferATQA[2];
byte bufferSize = sizeof(bufferATQA);

MFRC522 rfid(SS_PIN, RST_PIN);

void setup() {
  Serial.begin(9600); // Initialize serial for debugging
  DDRD &= ~_BV(PD2); // Set button pin to input
  PORTD |= _BV(PD2); // Set button pin pull-up
  DDRB |= _BV(PB1); // Set buzzer pin to output
  DDRD |= _BV(PD6); // Set GREEN led pin to output
  DDRD |= _BV(PD4); // Set RED led pin to output
  DDRD |= _BV(PD3); // Set BLUE led pin to output;
  DDRD |= _BV(PD7); // Set ultrasonic sensor trigger pin to output
  DDRB &= ~_BV(PB0); // Set ultrasonic sensor echo pin to input
  PORTD |= _BV(PD4); // Turn on RED led
  DDRC |= _BV(PC5); // Set relay pin to output
  PORTC &= ~_BV(PC5); // Lock door
  attachInterrupt(digitalPinToInterrupt(BUTTON_PIN), buttonInterrupt, CHANGE);  // Attach the interrupt
  SPI.begin(); // Begin SPI communication
  rfid.PCD_Init(); // Init RFID

}

// Button interrupt routine
void buttonInterrupt() {
  if (!buttonPressed && millis() - buttonLastStateTime > 50) {
    buttonPressed = true;
    lockState = true;
    buttonLastStateTime = millis(); // Keep track of when the button was last pushed
    Serial.println("Button pressed");
    PORTC |= _BV(PC5); // Unlock the door
    PORTD &= ~_BV(PD4); // Turn off RED led
  }
}

void alarmSetOff() {
  tone(BUZZER_PIN, 1500);
  delay(500);
  tone(BUZZER_PIN, 1000);
  delay(500);
  noTone(BUZZER_PIN);
}

void accessDenied() {
  tone(BUZZER_PIN, 1500);
  delay(200);
  noTone(BUZZER_PIN);
  delay(200);
  tone(BUZZER_PIN, 1500);
  delay(200);
  noTone(BUZZER_PIN);
  delay(200);
  tone(BUZZER_PIN, 1500);
  delay(200);
  noTone(BUZZER_PIN);
}

void resetRFID() {
  rfid.PICC_HaltA();
  rfid.PCD_StopCrypto1();
  rfid.PCD_Init();
}


void loop() {
  PORTD &= ~_BV(PD7); // Give a short LOW pulse on trigger to ensure clean pulse
  delayMicroseconds(5);
  PORTD |= _BV(PD7);
  delayMicroseconds(10);
  PORTD &= ~_BV(PD7);
  duration = pulseIn(ECHO_PIN, HIGH); // Read input from ECHO pin
  distance = (duration / 2) / 29.1; // Transform duration(time) into distance

  // 3 seconds delay after the button was pressed
  if (buttonPressed && millis() - buttonLastStateTime >= 3000) {
    if (distance <= 3) { // Door is shut
      buttonPressed = false;
      withValidKeycard = false;
      lockState = false;
      Serial.println("Button released");
      delay(1500); // Give some time for the door to close completely
      PORTC &= ~_BV(PC5); // Lock the door
      buttonLastStateTime = millis();
      resetRFID();
      PORTD |= _BV(PD4); // Turn on RED led
      PORTD &= ~_BV(PD6); // Turn off GREEN led
    }
  }

  // Door opened without keycard
  if (withValidKeycard == false && distance > 3) {
    alarmSetOff();
  }

  resetRFID();
  MFRC522::StatusCode result = rfid.PICC_WakeupA(bufferATQA, &bufferSize);
  rfid.PICC_RequestA(bufferATQA, &bufferSize);
  if (rfid.PICC_IsNewCardPresent()) { // New tag is available
    if (rfid.PICC_ReadCardSerial()) { // NUID has been readed

      if (memcmp(rfid.uid.uidByte, keyTagUID, sizeof(keyTagUID)) == 0) {
        Serial.println("Access is granted");
        withValidKeycard = true;
        PORTD &= ~_BV(PD4); // Turn off RED led
        PORTD |= _BV(PD6); // Turn on GREEN led
        PORTC |= _BV(PC5); // Unlock door
      }
      else if (lockState == true) {
        Serial.println("New Keycard Added");
        PORTD |= _BV(PD3); // Turn on BLUE led
        memcpy(keyTagUID, rfid.uid.uidByte, sizeof(keyTagUID));
        delay(3000);
        PORTD &= ~_BV(PD3); // Turn off BLUE led
      }
      else
      {
        Serial.println("Access denied");
        accessDenied();
      }
      resetRFID();
    }
  }

}
